//
// XMM Science Analysis System
//
// Giuseppe Vacanti (cosine) 
// October 2000
//
#include <Epproc.h>

#include <Dal.h>	// SAS
#include <OalServer.h>
#include <CalServer.h>
#include <param.h>
#include <errstr.h>
#include <metatask.h>
#include <Proc.h>
#include <Evaluator.h>

//AI: New call to epexposure and epreject
#include <epreject.h>
#include <epexposure.h>
#include <epframes.h>
#include <epevents.h>
#include <epauxcomb.h>
#include <badpix.h>
#include <dscopyblock.h>
#include <dsrm.h>
#include <dsrename.h>
#include <epnoise.h>
#include <epxrlcorr.h>
#include <epfast.h>


#include <algorithm>
#include <sstream>

Epproc::Epproc()
{
  PROC

    init();

  ENDPROC
}  


void Epproc::init_pimpl(){
  PROC
  
  ePS_ = new EpicProcState(12,
			   xmmPayLoad.instrument(Xmm::EPN),
			   Xmm::IMAGING);

  ePS_->allowed_modes.insert(std::make_pair(std::string("imaging"), Xmm::IMAGING));
  ePS_->allowed_modes.insert(std::make_pair(std::string("burst"), Xmm::BURST));
  ePS_->allowed_modes.insert(std::make_pair(std::string("timing"), Xmm::TIMING));
  
  ENDPROC
}

void Epproc::processExposure(Proc::ScienceExposure & exp){
  PROC

  errstr << "Processing ";
  ePS_->proc.print(exp, errstr);
  errstr << Sas::message(Msg::Meta, Msg::Silent);
  

  const std::list<const OdfSet*>& l = exp.exposure->sets();
  typedef std::list<const OdfSet*>::const_iterator LI;
  for(LI i=l.begin(); i!=l.end(); ++i){
    const OdfSet* os = (*i);
    if(find(ePS_->ccds.begin(), ePS_->ccds.end(), os->ccdNr()) != ePS_->ccds.end()){
      try {
    	  frames(os);
    	  epxrlcorr(os);
    	  badpixels(os);
    	  epreject(os);
    	  epnoise(os);	  
    	  events(os);
    	  attitude(os);
    	  finalize(os);
    	  filter(os);
      }
      catch(ProcException &e){
      }
    }
  }
  ENDPROC
}



void Epproc::frames(const OdfSet* os)
{
  PROC

  errstr << "Processing data set: " << os->name()
	 << Sas::message(Msg::Meta, Msg::Silent);

  oalServer << os->name();
  
  Info::Task* epframes = task("epframes");
  ParameterInfo& epframespars = *epframes->parameters();    
  epframespars.revertToDefault();
  epframespars["gtiset"]->setValue(ePS_->proc.framesgtiname());
  epframespars["eventset"]->setValue(ePS_->proc.frameseventsname());
  
  // Parameters transferred from the meta-task
  epframespars["mipthreshold"]->setValue(ePS_->mypars["mipthreshold"]);
  epframespars["mipmethod"]->setValue(ePS_->mypars["mipmethod"]);
  epframespars["mipdiscard"]->setValue(ePS_->mypars["mipdiscard"]);
  epframespars["mipthreshold"]->setValue(ePS_->mypars["mipthreshold"]);
  epframespars["mipdist"]->setValue(ePS_->mypars["mipdist"]);
  epframespars["wrongpixlimit"]->setValue(ePS_->mypars["wrongpixlimit"]);
  epframespars["guessdeltap"]->setValue(ePS_->mypars["guessdeltap"]);
  
  epframespars["set"]->setValue(os->name());
  if(os->dataMode() == Xmm::TIMING ||
     os->dataMode() ==Xmm::BURST){
    if(ePS_->mypars["withsrccoords"]->boolValue()){
      epframespars["withsrccoords"]->setValue(true);
      epframespars["srcra"]->setValue(ePS_->mypars["srcra"]);
      epframespars["srcdec"]->setValue(ePS_->mypars["srcdec"]);
    }
    if(os->dataMode() == Xmm::TIMING ||
       os->dataMode() == Xmm::BURST)
      epframespars["srcposition"]->setValue(ePS_->mypars["timingsrcposition"]);
  }
  

  if(!ePS_->proc.protectedrun(epframes))
    throw(ProcException(""));
  
  ePS_->ccd_gtis.push_back(ePS_->proc.framesgtiname());
  ePS_->epframes_Evlists.push_back(ePS_->proc.frameseventsname());

  // FIXME: make a copy to satisfy the i/f to badpixfind
  dataSetServer->copy(ePS_->proc.frameseventsname(),
		      ePS_->proc.badpixfindeventsname());
  

  ENDPROC
}


void Epproc::epxrlcorr(const OdfSet* os)
{
	PROC
    if(ePS_->mypars["runepxrlcorr"]->boolValue())
    {
    	if(os->dataMode() == Xmm::IMAGING)
    	{
    		oalServer << os->name();
    		Info::Task* epxrlcorr = task("epxrlcorr");
    		ParameterInfo& epxrlcorrpars = *epxrlcorr->parameters();

    		epxrlcorrpars.revertToDefault();
    		epxrlcorrpars["eventset"]->setValue(ePS_->proc.frameseventsname());
    		//epxrlcorrpars["outset"]->setValue(ePS_->proc.epxrlcorreventsname());
    		if(!ePS_->proc.protectedrun(epxrlcorr))
    		  throw(ProcException(""));
    	}
    	else
    	{
    	    errstr << "epxlcorr task is only available for" <<
    	      " maging mode. If you want to correct from X-ray loading "<<
    	      "in timing mode, please run epreject "
    		   << Sas::warning("Check Data mode");
    	}
    }
	ENDPROC
}


void Epproc::epreject(const OdfSet* os)
{
  PROC  

    if(ePS_->mypars["runepreject"]->boolValue())
      {
    	if(os->dataMode() == Xmm::TIMING ||
    			os->dataMode() == Xmm::COMPRESSED_TIMING ||
    			os->dataMode() == Xmm::BURST)
	  {
	    errstr << "Please be aware that the default values" <<
	      " for epreject may not be valid for timing observations"
		   << Sas::warning("Check epreject default values");	  	    
	  }


		Info::Task* epreject = task("epreject");
		ParameterInfo& eprejectpars = *epreject->parameters();

		eprejectpars.revertToDefault();

		eprejectpars["set"]->setValue(os->name());
		eprejectpars["eventset"]->setValue(ePS_->proc.frameseventsname());

		eprejectpars["badcolumnset"]->setValue(ePS_->mypars["badcolumnset"]);
		eprejectpars["sigma"]->setValue(ePS_->mypars["sigma"]);

		if (ePS_->mypars["runepnoise"]->boolValue())
		{
			eprejectpars["withnoisehandling"]->setValue(false);
			errstr << "noise reduction already applied in epnoise."
					<<" epreject will not apply noise reduction correction."
					<< Sas::warning("NoiseCorruectionAlreadyApplied");
		}
		else
			eprejectpars["withnoisehandling"]->setValue(ePS_->mypars["withnoisehandling"]);

		eprejectpars["noiseparameters"]->setValue(ePS_->mypars["noiseparameters"]);
		eprejectpars["withoffsetmap"]->setValue(ePS_->mypars["withoffsetmap"]);
		eprejectpars["withxrlcorrection"]->setValue(ePS_->mypars["withxrlcorrection"]);
		eprejectpars["withsoftflarescreening"]->setValue(ePS_->mypars["withsoftflarescreening"]);
		eprejectpars["softflarethreshold1"]->setValue(ePS_->mypars["softflarethreshold1"]);
		eprejectpars["softflarethreshold2"]->setValue(ePS_->mypars["softflarethreshold2"]);
		eprejectpars["softflaresmooth"]->setValue(ePS_->mypars["softflaresmooth"]);
		eprejectpars["softflareenergyrange"]->setValue(ePS_->mypars["softflareenergyrange"]);
		eprejectpars["softflaresmoothparams"]->setValue(ePS_->mypars["softflaresmoothparams"]);

		if(!ePS_->proc.protectedrun(epreject))
		  throw(ProcException(""));
	
      }

  ENDPROC
}




void Epproc::epnoise(const OdfSet* os)
{
  PROC
    if( (ePS_->mypars["runepnoise"]->boolValue()))
    {
	  if (os->dataMode() != Xmm::IMAGING)
		  Sas::error("IncorrectMode");

		  Info::Task* epnoise = task("epnoise");
		  ParameterInfo& epnoisepar = *epnoise->parameters();

		  epnoisepar.revertToDefault();

		  std::vector<std::string> epnoiseFiles;
		  epnoiseFiles.push_back(ePS_->proc.frameseventsname());
		  epnoisepar["set"]->setValue(epnoiseFiles);
		  epnoisepar["identifynoisyframes"]->setValue(true);
		  epnoisepar["applyfilter"]->setValue(false);
		  epnoisepar["noisecut"]->setValue(ePS_->mypars["noisecut"]);
		  epnoisepar["sigmacut"]->setValue(ePS_->mypars["sigmacut"]);
		  epnoisepar["savemasks"]->setValue(ePS_->mypars["savemasks"]);
		  if(!ePS_->proc.protectedrun(epnoise))
			  throw(ProcException(""));
    }
  ENDPROC
}



bool Epproc::couldRunBadpixfindOn(const OdfSet *)
{
  PROC  RETURN true;
}

void Epproc::badpix(const OdfSet* os)
{
  PROC
  Info::Task* badpix = task("badpix");
  ParameterInfo& badpixpars = *badpix->parameters();
  badpixpars.revertToDefault();
  badpixpars["badpixset"]->setValue(ePS_->proc.badpixfindname());
  badpixpars["eventset"]->setValue(ePS_->proc.frameseventsname());

  if(os->dataMode() == Xmm::TIMING ||
     os->dataMode() == Xmm::BURST){
    badpixpars["getuplnkbadpix"]->setValue(false);
    badpixpars["getotherbadpix"]->setValue(false);
    badpixpars["getnewbadpix"]->setValue(false);
    badpixpars["emptyextension"]->setValue(true);
  } else {
    badpixpars["getuplnkbadpix"]->setValue(true);
    badpixpars["getotherbadpix"]->setValue(true);
    badpixpars["getnewbadpix"]->setValue(ePS_->has_badpixset);
    badpixpars["emptyextension"]->setValue(false);
  }

  if(!ePS_->proc.protectedrun(badpix))
    throw(ProcException(""));
  
  ENDPROC
}

void Epproc::events(const OdfSet* os)
{
  PROC
    Info::Task* epevents = task("epevents");
  ParameterInfo& epeventspars = *epevents->parameters();
  epeventspars.revertToDefault();
  epeventspars["randomizeposition"]->setValue(ePS_->mypars["randomizeposition"]);
  epeventspars["randomizeenergy"]->setValue(ePS_->mypars["randomizeenergy"]);
  epeventspars["testenergywidth"]->setValue(ePS_->mypars["testenergywidth"]);
  epeventspars["gainctiaccuracy"]->setValue(ePS_->mypars["gainctiaccuracy"]);
  epeventspars["reemissionthresh"]->setValue(ePS_->mypars["reemissionthresh"]);
  epeventspars["mappatterntype"]->setValue(ePS_->mypars["mappatterntype"]);
  epeventspars["patternanalysis"]->setValue(ePS_->mypars["patternanalysis"]);
  epeventspars["withtempcorrection"]->setValue(ePS_->mypars["withtempcorrection"]);
  epeventspars["withframecti"]->setValue(ePS_->mypars["withframecti"]);
  epeventspars["withrdpha"]->setValue(ePS_->mypars["withrdpha"]);
  epeventspars["checksasmip"]->setValue(ePS_->mypars["checksasmip"]);
  epeventspars["rdphatimebinsize"]->setValue(ePS_->mypars["rdphatimebinsize"]);


  if(os->dataMode() == Xmm::IMAGING)
    epeventspars["withoutoftime"]->setValue(ePS_->mypars["withoutoftime"]);
  else
    epeventspars["withoutoftime"]->setValue(false);

  epeventspars["eventset"]->setValue(ePS_->proc.frameseventsname());
  epeventspars["outset"]->setValue(ePS_->proc.eventsname());
  if(!ePS_->proc.protectedrun(epevents))
    throw(ProcException(""));
  ePS_->exposure_evlists.push_back(ePS_->proc.eventsname());
  if (ePS_->mypars["pileuptempfile"])
	  dataSetServer->copy(ePS_->proc.eventsname(),ePS_->proc.pileupname());

  ENDPROC

}

void Epproc::finalizeHook(const OdfSet *){
  PROC

  // Also rename the EXPOSURE extension to EXPOSUnn, so the
  // eveselect can update the exposure information, and the ccd
  // level event list can be used as is.
  //
  Info::Task* dsrename = task("dsrename");
  ParameterInfo& dsrenamepars = *dsrename->parameters();
  dsrenamepars.revertToDefault();
  dsrenamepars["object"]->setValue(ePS_->proc.eventsname() +
				   ":EXPOSURE");
  
  std::stringstream sstr;
  sstr << "EXPOSU" << std::setw(2) << std::setfill('0')
       << oalServer.ccd();
  dsrenamepars["newname"]->setValue(sstr.str());
  if(!ePS_->proc.protectedrun(dsrename))
    throw(ProcException(""));



  // Remove a number of columns from the exposure extension that
  // are not used.
  if(ePS_->mypars["deleteexposurecolumns"]->boolValue()){
	  std::string s(ePS_->proc.eventsname() + ":" + sstr.str());
	  std::vector<std::string> ss(5,s);
    ss[0] += ":FRAMELIM";
    ss[1] += ":NDSCLIN1";
    ss[2] += ":NDSCLIN2";
    ss[3] += ":NDSCLIN3";
    ss[4] += ":NDSCLIN4";
    Info::Task *dsrm = task("dsrm");
    ParameterInfo& dsrmpars = *dsrm->parameters();
    dsrmpars.revertToDefault();
    dsrmpars["objects"]->setValue(ss);
    if(!ePS_->proc.protectedrun(dsrm))
      Sas::message(Msg::Meta, Msg::Sparse, "Could not remove extra columns from " + s);       
  }
  
  ENDPROC
}


void Epproc::epfast(const Proc::ScienceExposure& thisexp)
{
	PROC

	if ( thisexp.mode == Xmm::TIMING ||
			thisexp.mode == Xmm::BURST )
	{
		if (!ePS_->mypars["withrdpha"]->boolValue() )
		{
			Info::Task* epfast = task("epfast");
			ParameterInfo& epfastpars = *epfast->parameters();

			epfastpars.revertToDefault();
			epfastpars["evfile"]->setValue(ePS_->proc.timingeventsname());
			if(!ePS_->proc.protectedrun(epfast))
				throw(ProcException(""));
		}
		else
		{
			errstr << "rate-dependent PHA correction already applied in epevents."
					<<" epfast will not run."
					<< Sas::warning("PHACorrectionAlreadyApplied");
		}
	}
	else
	{
		errstr << "epfast task is only available for" <<
				" fast modes. "
				<< Sas::warning("ModeNotAvailable");
	}

	ENDPROC
}



void Epproc::screenExposure(const Proc::ScienceExposure& thisexp)
{
  PROC

  /*
   * run epfast first in case we are in TI or BU and
   */
  if(ePS_->mypars["runepfast"]->boolValue())
	  	  epfast(thisexp);



  ePS_->proc.print(thisexp, errstr);

  std::vector<std::string> eventlist;
  
  if(ePS_->mypars["runepexposure"]->boolValue())
  {
	  Info::Task* epexposure = task("epexposure");
	  ParameterInfo& epexposurepars = *epexposure->parameters();

	  epexposurepars.revertToDefault();

	  if(thisexp.mode == Xmm::IMAGING ||
			  thisexp.mode == Xmm::REDUCED_IMAGING)
		  eventlist.push_back(ePS_->proc.imagingeventsname());
	  else if(thisexp.mode == Xmm::TIMING ||
			  thisexp.mode == Xmm::COMPRESSED_TIMING ||
			  thisexp.mode == Xmm::BURST)
		  eventlist.push_back(ePS_->proc.timingeventsname());

	  epexposurepars["eventsets"]->setValue(eventlist);
	  epexposurepars["randomizetime"]->setValue(ePS_->mypars["randomizetime"]->boolValue());
	  epexposurepars["spatialexposure"]->setValue(ePS_->mypars["spatialexposure"]->boolValue());

	  if(!ePS_->proc.protectedrun(epexposure))
		  throw(ProcException(""));

      }


  ENDPROC
}

void Epproc::combineHook(ParameterInfo& evlistcombpars)
{
  PROC
    // Change the columns that are propagates by evlistcomb
    // 
  std::vector<std::string> epntimcolnames, epntimcoltypes;
      epntimcolnames.push_back("TIME");    epntimcoltypes.push_back("double");
      epntimcolnames.push_back("RAWX");    epntimcoltypes.push_back("int16"); 
      epntimcolnames.push_back("RAWY");    epntimcoltypes.push_back("int16"); 
      // epntimcolnames.push_back("DETX");    epntimcoltypes.push_back("int16");
      // epntimcolnames.push_back("DETY");    epntimcoltypes.push_back("int16");
      epntimcolnames.push_back("PHA");     epntimcoltypes.push_back("int16");
      epntimcolnames.push_back("PI");      epntimcoltypes.push_back("int16");
      epntimcolnames.push_back("FLAG");    epntimcoltypes.push_back("int32");
      epntimcolnames.push_back("PATTERN"); epntimcoltypes.push_back("int8");
      epntimcolnames.push_back("PAT_ID");  epntimcoltypes.push_back("int16"); 
      epntimcolnames.push_back("OFFSETX");  epntimcoltypes.push_back("int16"); 
      // epntimcolnames.push_back("CCDNR");   epntimcoltypes.push_back("int8");

      evlistcombpars["epntimcolnames"]->setValue(epntimcolnames);
      evlistcombpars["epntimcoltypes"]->setValue(epntimcoltypes);
      std::vector<std::string> mainAttributes;
      if(ePS_->mypars["runepxrlcorr"]->boolValue())
      {
    	  mainAttributes.push_back("XRL_CORR");
      }

      if(ePS_->mypars["withrdpha"]->boolValue())
      {
    	  mainAttributes.push_back("PHA_RDCO");
      }


      if(ePS_->mypars["withxrlcorrection"]->boolValue())
      {
    	  mainAttributes.push_back("REJ_CORR");
      }


    	  evlistcombpars["mainattributes"]->setValue(mainAttributes);

  ENDPROC

}

